﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

namespace HeuristicLab.Algorithms.DataAnalysis.Symbolic {
  [StorableClass]
  [Item("IntervalInterpreter", "Intperter for calculation of intervals of symbolic models.")]
  public sealed class IntervalInterpreter : ParameterizedNamedItem, IStatefulItem {

    private const string EvaluatedSolutionsParameterName = "EvaluatedSolutions";
    private static int InstructionCount = 0;

    public IFixedValueParameter<IntValue> EvaluatedSolutionsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[EvaluatedSolutionsParameterName]; }
    }

    public int EvaluatedSolutions {
      get { return EvaluatedSolutionsParameter.Value.Value; }
      set { EvaluatedSolutionsParameter.Value.Value = value; }
    }

    private IntervalInterpreter(bool deserializing) : base(deserializing) { }
    private IntervalInterpreter(IntervalInterpreter original, Cloner cloner)
        : base(original, cloner) { }

    public IntervalInterpreter()
        : base("IntervalInterpreter", "Intperter for calculation of intervals of symbolic models.") { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new IntervalInterpreter(this, cloner);
    }

    private readonly object syncRoot = new object();

    #region IStatefulItem Members
    public void InitializeState() {
      EvaluatedSolutions = 0;
    }
    public void ClearState() { }
    #endregion

    private static void ResetInstrucitonCount() {
      InstructionCount = 0;
    }


    public Interval GetSymbolicExressionTreeIntervals(ISymbolicExpressionTree tree, IDataset dataset, IEnumerable<int> rows = null) {
      lock (syncRoot) {
        EvaluatedSolutions++;
        ResetInstrucitonCount();
      }
      var instructions = PrepareInterpreterState(tree, rows, dataset);
      var x = Evaluate(instructions);

      return x;
    }

    public Interval GetSymbolicExressionTreeIntervals(ISymbolicExpressionTree tree, Dictionary<string, Interval> customIntervals, IEnumerable<int> rows = null) {
      lock (syncRoot) {
        EvaluatedSolutions++;
        ResetInstrucitonCount();
      }
      var instructions = PrepareInterpreterState(tree, rows, null, customIntervals);
      var x = Evaluate(instructions);

      return x;
    }

    public Interval GetSymbolicExressionTreeIntervals(ISymbolicExpressionTree tree, IDataset dataset,
      Dictionary<string, Interval> customIntervals, out Dictionary<ISymbolicExpressionTreeNode, Interval> intervals, IEnumerable<int> rows = null) {
      lock (syncRoot) {
        EvaluatedSolutions++;
        ResetInstrucitonCount();
      }
      intervals = new Dictionary<ISymbolicExpressionTreeNode, Interval>();
      var instructions = PrepareInterpreterState(tree, rows, dataset, customIntervals);
      var x = Evaluate(instructions, intervals);

      return x;
    }

    public Interval GetSymbolicExressionTreeIntervals(ISymbolicExpressionTree tree, IDataset dataset,
      out Dictionary<ISymbolicExpressionTreeNode, Interval> intervals, IEnumerable<int> rows = null) {
      lock(syncRoot) {
        EvaluatedSolutions++;
        ResetInstrucitonCount();
      }
      intervals = new Dictionary<ISymbolicExpressionTreeNode, Interval>();
      var instructions = PrepareInterpreterState(tree, rows, dataset);
      var x = Evaluate(instructions, intervals);

      return x;
    }

    public Interval GetSymbolicExressionTreeIntervals(ISymbolicExpressionTree tree, Dictionary<string, Interval> customIntervals, 
      out Dictionary<ISymbolicExpressionTreeNode, Interval> intervals, IEnumerable<int> rows = null) {
      lock (syncRoot) {
        EvaluatedSolutions++;
        ResetInstrucitonCount();
      }
      intervals = new Dictionary<ISymbolicExpressionTreeNode, Interval>();
      var instructions = PrepareInterpreterState(tree, rows, null, customIntervals);
      var x = Evaluate(instructions, intervals);

      return x;
    }

    private static Instruction[] PrepareInterpreterState(ISymbolicExpressionTree tree,
      IEnumerable<int> rows = null, IDataset dataset = null, Dictionary<string, Interval> customIntervals = null) {
      Instruction[] code = SymbolicExpressionTreeCompiler.Compile(tree, OpCodes.MapSymbolToOpCode);

      if (dataset == null && customIntervals == null)
        throw new Exception("No dataset or ranges for intervals are given!");

      if(rows == null)
        rows = Enumerable.Range(0, dataset.Rows);

      foreach (Instruction instr in code) {
        if (instr.opCode == OpCodes.Variable) {
          var variableTreeNode = (VariableTreeNode)instr.dynamicNode;
          IList<double> values = new List<double>();

          if (customIntervals != null) {
            if (customIntervals.ContainsKey(variableTreeNode.VariableName)) {
              instr.data = customIntervals[variableTreeNode.VariableName];
            }
          } else {
            foreach (var rowEnum in rows) {
              values.Add(dataset.GetReadOnlyDoubleValues(variableTreeNode.VariableName)[rowEnum]);
            }
            instr.data = new Interval(values.Min(), values.Max());
          }
        }
      }
      return code;
    }

    private Interval Evaluate(Instruction[] instructions, Dictionary<ISymbolicExpressionTreeNode, Interval> intervals = null) {
      Instruction currentInstr = instructions[InstructionCount++];
      Interval intermediate = null;

      switch (currentInstr.opCode) {
        //Elementary arithmetic rules
        case OpCodes.Add: {
            intermediate = Evaluate(instructions, intervals);
            for (int i = 1; i < currentInstr.nArguments; i++) {
              intermediate = Interval.Add(intermediate, Evaluate(instructions, intervals));
            }
            break;
          }
        case OpCodes.Sub: {
            intermediate = Evaluate(instructions, intervals);
            for (int i = 1; i < currentInstr.nArguments; i++) {
              intermediate = Interval.Subtract(intermediate, Evaluate(instructions, intervals));
            }
            break;
          }
        case OpCodes.Mul: {
            intermediate = Evaluate(instructions, intervals);
            for (int i = 1; i < currentInstr.nArguments; i++) {
              intermediate = Interval.Multiply(intermediate, Evaluate(instructions, intervals));
            }
            break;
          }
        case OpCodes.Div: {
            intermediate = Evaluate(instructions, intervals);
            for (int i = 1; i < currentInstr.nArguments; i++) {
              intermediate = Interval.Divide(intermediate, Evaluate(instructions, intervals));
            }
            break;
          }
        //Trigonometric functions
        case OpCodes.Sin: {
            intermediate = Interval.Sine(Evaluate(instructions, intervals));
            break;
          }
        case OpCodes.Cos: {
            intermediate = Interval.Cosine(Evaluate(instructions, intervals));
            break;
          }
        case OpCodes.Tan: {
            intermediate = Interval.Tangens(Evaluate(instructions, intervals));
            break;
          }
        //Exponential functions
        case OpCodes.Log: {
            intermediate = Interval.Logarithm(Evaluate(instructions, intervals));
            break;
          }
        case OpCodes.Exp: {
            intermediate = Interval.Exponential(Evaluate(instructions, intervals));
            break;
          }
        case OpCodes.Power: {
            intermediate = Evaluate(instructions, intervals);
            for (int i = 1; i < currentInstr.nArguments; i++) {
              intermediate = Interval.Power(intermediate, Evaluate(instructions, intervals));
            }
            break;
          }
        case OpCodes.Square: {
            intermediate = Interval.Square(Evaluate(instructions, intervals));
            break;
          }
        case OpCodes.Root: {
            intermediate = Evaluate(instructions, intervals);
            for (int i = 1; i < currentInstr.nArguments; i++) {
              intermediate = Interval.Root(intermediate, Evaluate(instructions, intervals));
            }
            break;
          }
        case OpCodes.SquareRoot: {
            intermediate = Interval.SquareRoot(Evaluate(instructions, intervals));
            break;
          }
        //Variables, Constants, ...
        case OpCodes.Variable: {
            intervals.Add(currentInstr.dynamicNode, (Interval)currentInstr.data);
            return (Interval)currentInstr.data;
          }
        case OpCodes.Constant: {
            var constTreeNode = (ConstantTreeNode)currentInstr.dynamicNode;
            var inter = new Interval(constTreeNode.Value, constTreeNode.Value);
            intervals.Add(currentInstr.dynamicNode, inter);
            return inter;
          }
        default:
          throw new NotSupportedException("Tree contains an unknown symbol.");
      }

      if (intervals != null)
        intervals.Add(currentInstr.dynamicNode, intermediate);
      return intermediate;
    }

    public static bool IsCompatible(ISymbolicExpressionTree tree) {
      var containsUnknownSyumbol = (
        from n in tree.Root.GetSubtree(0).IterateNodesPrefix()
        where
          !(n.Symbol is StartSymbol) &&
          !(n.Symbol is Addition) &&
          !(n.Symbol is Subtraction) &&
          !(n.Symbol is Multiplication) &&
          !(n.Symbol is Division) &&
          !(n.Symbol is Sine) &&
          !(n.Symbol is Cosine) &&
          !(n.Symbol is Tangent) &&
          !(n.Symbol is Logarithm) &&
          !(n.Symbol is Exponential) &&
          !(n.Symbol is Power) &&
          !(n.Symbol is Square) &&
          !(n.Symbol is Root) &&
          !(n.Symbol is SquareRoot) &&
          !(n.Symbol is Problems.DataAnalysis.Symbolic.Variable) &&
          !(n.Symbol is Constant)
        select n).Any();
      return !containsUnknownSyumbol;
    }
  }
}
